"""Stochastic Lanczos quadrature."""

from matfree import decomp, montecarlo
from matfree.backend import func, linalg, np


def logdet(*args, **kwargs):
    """Estimate the log-determinant of a matrix."""
    return trace_of_matfun(np.log, *args, **kwargs)


# todo: nuclear norm, schatten-p norms.
#  But for this we should use bi-diagonalisation


def trace_of_matfun(matfun, Av, order, /, **kwargs):
    """Compute the trace of the function of a matrix."""
    quadratic_form = quadratic_form_slq(matfun, Av, order)
    return montecarlo.estimate(quadratic_form, **kwargs)


def quadratic_form_slq(matfun, Av, order, /):
    """Approximate quadratic form for stochastic Lanczos quadrature."""

    def quadform(v0, /):
        algorithm = decomp.lanczos(order)
        _, tridiag = decomp.decompose_fori_loop(0, order + 1, Av, v0, alg=algorithm)
        (diag, off_diag) = tridiag

        # todo: once jax supports eigh_tridiagonal(eigvals_only=False),
        #  use it here. Until then: an eigen-decomposition of size (order + 1)
        #  does not hurt too much...
        diag = linalg.diagonal_matrix(diag)
        offdiag1 = linalg.diagonal_matrix(off_diag, -1)
        offdiag2 = linalg.diagonal_matrix(off_diag, 1)
        dense_matrix = diag + offdiag1 + offdiag2
        eigvals, eigvecs = linalg.eigh(dense_matrix)

        # Since Q orthogonal (orthonormal) to v0, Q v = Q[0],
        # and therefore (Q v)^T f(D) (Qv) = Q[0] * f(diag) * Q[0]
        (dim,) = v0.shape

        fx_eigvals = func.vmap(matfun)(eigvals)
        return dim * linalg.vecdot(eigvecs[0, :], fx_eigvals * eigvecs[0, :])

    return quadform



"""NumPy-style API."""

# In here, we loosely follow the Array API:
#
# https://data-apis.org/array-api/2022.12/
#
# But deviate in a few points:
# * The functions here do not have all the arguments specified in the API
#   (we only wrap the arguments we need)
# * Our current version of diag/diagonal is slightly different
# * We do not use methods on Array types, e.g. shape(), dtype(). Instead
#   these are functions. (Not all backends might always follow this method interface.)
# * We do not implement any constants (e.g. NaN, Pi). Instead, these are methods.


import jax.numpy as jnp

# Creation functions:


def arange(start, /, stop=None, step=1):
    return jnp.arange(start, stop, step)


def asarray(obj, /):
    return jnp.asarray(obj)


def eye(n_rows):
    return jnp.eye(n_rows)


def ones_like(x, /):
    return jnp.ones_like(x)


def zeros(shape, *, dtype=None):
    return jnp.zeros(shape, dtype=dtype)


# Element-wise functions


def abs(x, /):  # noqa: A001
    return jnp.abs(x)


def log(x, /):
    return jnp.log(x)


def isnan(x, /):
    return jnp.isnan(x)


def sin(x, /):
    return jnp.sin(x)


def cos(x, /):
    return jnp.cos(x)


def sqrt(x, /):
    return jnp.sqrt(x)


# Utility functions


def any(x, /):  # noqa: A001
    return jnp.any(x)


def allclose(x1, x2, /, *, rtol=1e-5, atol=1e-8):
    return jnp.allclose(x1, x2, rtol=rtol, atol=atol)


# Statistical functions


def mean(x, /, axis=None):
    return jnp.mean(x, axis)


def std(x, /, axis=None):
    return jnp.std(x, axis)


def sum(x, /, axis=None):  # noqa: A001
    return jnp.sum(x, axis)


def maximum(x1, x2):  # todo: call max()
    return jnp.maximum(x1, x2)


def nanmean(x, /, axis=None):
    return jnp.nanmean(x, axis)


# Searching functions


def where(condition, x1, x2, /):
    return jnp.where(condition, x1, x2)


# Manipulation functions


def reshape(x, /, shape):
    return jnp.reshape(x, shape)


def flip(x, /):
    return jnp.flip(x)


def roll(x, /, shift):
    return jnp.roll(x, shift)


# Functional implementation of what are usually array-methods


def shape(x, /):
    return jnp.shape(x)


def dtype(x, /):
    return jnp.dtype(x)


# Functional implementation of constants


def nan():
    return jnp.nan



"""Control_variates benchmark.

Runtime: ~10 seconds.
"""

from matfree import hutch, montecarlo
from matfree.backend import func, linalg, np, plt, prng, progressbar, time


def rmse_relative(received, *, expected):
    """Compute the relative root-mean-square error."""
    return linalg.vector_norm((received - expected) / expected) / np.sqrt(expected.size)


def error_and_time(fun, error_fun):
    """Compute error and runtime of a function with a single outputs."""

    def fun_wrapped(*args, **kwargs):
        # Execute once for compilation
        _ = fun(*args, **kwargs)

        # Execute and time
        t0 = time.perf_counter()
        result = fun(*args, **kwargs)
        result.block_until_ready()
        t1 = time.perf_counter()
        return error_fun(result), (t1 - t0)

    return fun_wrapped


def problem(n):
    """Create an example problem."""

    # This function has a Jacobian with x-shaped sparsity pattern
    # We expect control variates to do pretty well
    # (But I don't know why)
    def f(x):
        return np.sin(np.roll(np.sin(np.flip(np.cos(x)) + 1) ** 2, 1)) * np.sin(x**2)

    key = prng.prng_key(seed=2)
    x0 = prng.uniform(key, shape=(n,))

    _, jvp = func.linearize(f, x0)
    J = func.jacfwd(f)(x0)
    trace = linalg.trace(J)
    sample_fun = montecarlo.normal(shape=(n,), dtype=float)

    return (jvp, trace, J), (key, sample_fun)


if __name__ == "__main__":
    dim = 100
    num_samples = 2 ** np.arange(4, 10)
    num_restarts = 5

    (Av, trace, J), (k, sample_fun) = problem(dim)

    error_fun = func.partial(rmse_relative, expected=trace)

    @func.partial(error_and_time, error_fun=error_fun)
    @func.partial(func.jit, static_argnums=0)
    def fun1(num, key):
        """Estimate the trace conventionally."""
        return hutch.trace(
            Av, key=key, sample_fun=sample_fun, num_batches=num, num_samples_per_batch=1
        )

    @func.partial(error_and_time, error_fun=error_fun)
    @func.partial(func.jit, static_argnums=0)
    def fun2(num, key):
        """Estimate trace and diagonal jointly and discard the diagonal."""
        trace2, _ = hutch.trace_and_diagonal(
            Av, key=key, num_levels=num, sample_fun=sample_fun
        )
        return trace2

    errors1, stds1, times1 = [], [], []
    errors2, stds2, times2 = [], [], []

    for n in progressbar.progressbar(num_samples):
        test_keys = prng.split(k, num=num_restarts)
        e1, t1 = zip(*[fun1(int(n), ke) for ke in test_keys])
        e2, t2 = zip(*[fun2(int(n), ke) for ke in test_keys])

        errors1.append(np.mean(np.asarray(e1)))
        stds1.append(np.std(np.asarray(e1)))
        times1.append(min(t1))

        errors2.append(np.mean(np.asarray(e2)))
        stds2.append(np.std(np.asarray(e2)))
        times2.append(min(t2))

    errors1 = np.asarray(errors1)
    stds1 = np.asarray(stds1)
    times1 = np.asarray(times1)

    errors2 = np.asarray(errors2)
    stds2 = np.asarray(stds2)
    times2 = np.asarray(times2)

    fig, axes = plt.subplots(ncols=2, tight_layout=True, figsize=(10, 5), dpi=100)

    ax_wp, ax_sparsity = axes
    ax_wp.set_title("Trace estimation")
    ax_wp.loglog(
        times1,
        errors1,
        "o-",
        markeredgecolor="k",
        label="Conventional trace estimation",
    )

    ax_wp.fill_between(
        times1,
        errors1 - stds1,  # type: ignore
        errors1 + stds1,  # type: ignore
        alpha=0.25,
    )
    ax_wp.loglog(
        times2,
        errors2,
        "^-",
        markeredgecolor="k",
        label="Joint trace and diagonal estimation",
    )
    ax_wp.fill_between(
        times2,
        errors2 - stds2,  # type: ignore
        errors2 + stds2,  # type: ignore
        alpha=0.25,
    )
    ax_wp.set_ylabel(f"Relative error (mean & std, {num_restarts} restarts)")
    ax_wp.set_xlabel(f"Wall clock time (minimum, {num_restarts} restarts)")
    ax_wp.legend()

    ax_sparsity.set_title("Sparsity pattern of Jacobian")
    ax_sparsity.spy(J)
    ax_sparsity.set_xticks(())
    ax_sparsity.set_yticks(())
    plt.show()



"""Plotting functionality."""

# Not part of dependencies because not used by default, only for debugging.
import matplotlib.pyplot  # noqa: ICN001


def subplots(nrows=1, ncols=1, *, figsize, tight_layout=True, dpi=100):
    return matplotlib.pyplot.subplots(
        nrows=nrows, ncols=ncols, figsize=figsize, tight_layout=tight_layout, dpi=dpi
    )


def show():
    return matplotlib.pyplot.show()



"""Function transformations (algorithmic differentiation, vmap, partial, and so on)."""

# API-wise, we tend to follow JAX and functools.
# But we only implement those function arguments that we need.
# For example, algorithmic differentiation functions do not offer a 'has_aux' argument
# at the moment, because we don't need it.

import functools

import jax

# Vectorisation


def vmap(fun, /, in_axes=0, out_axes=0):
    return jax.vmap(fun, in_axes=in_axes, out_axes=out_axes)


# Partial and the like


def partial(func, /, *args, **kwargs):
    return functools.partial(func, *args, **kwargs)


# Algorithmic differentiation


def linearize(func, /, *args):
    return jax.linearize(func, *args)


def jacfwd(fun, /, argnums=0):
    return jax.jacfwd(fun, argnums)


# Inferring input and output shapes:


def eval_shape(func, /, *args, **kwargs):
    return jax.eval_shape(func, *args, **kwargs)


# Compilation (don't use in source!)


def jit(fun, /, *, static_argnums=None, static_argnames=None):
    return jax.jit(fun, static_argnames=static_argnames, static_argnums=static_argnums)



"""Test-utilities."""

import jax.test_util
import pytest
import pytest_cases


def fixture(name=None):
    return pytest_cases.fixture(name=name)


def parametrize(argnames, argvalues, /):
    return pytest.mark.parametrize(argnames, argvalues)


def check_grads(fun, /, args, *, order, atol, rtol):
    return jax.test_util.check_grads(fun, args, order=order, atol=atol, rtol=rtol)


def raises(err, /):
    return pytest.raises(err)



"""Monte-Carlo estimation."""

from matfree.backend import containers, control_flow, func, np, prng
from matfree.backend.typing import Any, Array, Callable


def estimate(
    fun: Callable,
    /,
    *,
    key: Array,
    sample_fun: Callable,
    num_batches: int = 1,
    num_samples_per_batch: int = 10_000,
) -> Array:
    """Monte-Carlo estimation: Compute the expected value of a function.

    Parameters
    ----------
    fun:
        Function whose expected value shall be estimated.
    key:
        Pseudo-random number generator key.
    sample_fun:
        Sampling function.
        For trace-estimation, use
        either [montecarlo.normal(...)][matfree.montecarlo.normal]
        or [montecarlo.rademacher(...)][matfree.montecarlo.normal].
    num_batches:
        Number of batches when computing arithmetic means.
    num_samples_per_batch:
        Number of samples per batch.
    """
    fun_mc = _montecarlo(fun, sample_fun=sample_fun)
    fun_single_batch = _mean_vmap(fun_mc, num_samples_per_batch)
    fun_batched = _mean_loop(fun_single_batch, num_batches)
    return fun_batched(key)


def _montecarlo(f, /, *, sample_fun):
    """Turn a function into a Monte-Carlo problem.

    More specifically, f(x) becomes g(key) = f(h(key)),
    using a sample function h: key -> x.
    This can then be evaluated and averaged in batches, loops, and compositions thereof.
    """
    # todo: what about randomised QMC? How do we best implement this?

    def f_mc(key, /):
        sample = sample_fun(key)
        return f(sample)

    return f_mc


def _mean_vmap(f, num, /):
    """Compute a batch-mean via jax.vmap."""

    def f_mean(key, /):
        subkeys = prng.split(key, num)
        fx_values = func.vmap(f)(subkeys)
        return np.nanmean(fx_values, axis=0)

    return f_mean


class _MState(containers.NamedTuple):
    mean: Any
    key: Any


def _mean_loop(f, num, /):
    """Compute an on-the-fly mean via a for-loop."""

    def f_mean(key, /):
        # Initialise
        fx_shape = func.eval_shape(f, key).shape
        mean = np.zeros(fx_shape, dtype=float)
        mstate = _MState(mean=mean, key=key)

        # Run for-loop
        increment = func.partial(_mean_increment, fun=f)
        mstate = control_flow.fori_loop(0, num, body_fun=increment, init_val=mstate)
        mean, _key = mstate  # todo: why not return key?

        # Return results
        return mean

    return f_mean


def _mean_increment(i, mstate: _MState, fun) -> _MState:
    """Increment the current mean-estimate."""
    # Read and split key
    mean, key = mstate
    _, subkey = prng.split(key)

    # Evaluate function
    fx_values = fun(subkey)

    # Update mean estimate
    mean_new = np.sum(np.asarray([mean * i, fx_values]), axis=0) / (i + 1)
    return _MState(mean=mean_new, key=subkey)


def normal(*, shape, dtype=float):
    """Construct a function that samples from a standard normal distribution."""

    def fun(key):
        return prng.normal(key, shape=shape, dtype=dtype)

    return fun


def rademacher(*, shape, dtype=float):
    """Construct a function that samples from a Rademacher distribution."""

    def fun(key):
        return prng.rademacher(key, shape=shape, dtype=dtype)

    return fun


class _VDCState(containers.NamedTuple):
    n: int
    vdc: float
    denom: int


def van_der_corput(n, /, base=2):
    """Compute the 'n'th element of the Van-der-Corput sequence."""
    state = _VDCState(n, vdc=0, denom=1)

    vdc_modify = func.partial(_van_der_corput_modify, base=base)
    state = control_flow.while_loop(_van_der_corput_cond, vdc_modify, state)
    return state.vdc


def _van_der_corput_cond(state: _VDCState):
    return state.n > 0


def _van_der_corput_modify(state: _VDCState, *, base):
    denom = state.denom * base
    num, remainder = divmod(state.n, base)
    vdc = state.vdc + remainder / denom
    return _VDCState(num, vdc, denom)



"""Pseudo-random-number utilities."""


import jax.random


def prng_key(seed):
    return jax.random.PRNGKey(seed=seed)


def split(key, num=2):
    return jax.random.split(key, num=num)


def normal(key, *, shape, dtype=None):
    if dtype is None:
        return jax.random.normal(key, shape=shape)
    return jax.random.normal(key, shape=shape, dtype=dtype)


def uniform(key, *, shape, dtype=None):
    if dtype is None:
        return jax.random.uniform(key, shape=shape)
    return jax.random.uniform(key, shape=shape, dtype=dtype)


def rademacher(key, *, shape, dtype=None):
    if dtype is None:
        return jax.random.rademacher(key, shape=shape)
    return jax.random.rademacher(key, shape=shape, dtype=dtype)



"""Matrix decomposition algorithms."""

from matfree.backend import containers, control_flow, func, linalg, np
from matfree.backend.typing import Any, Array, Callable


class DecompAlg(containers.NamedTuple):
    """Matrix decomposition algorithm."""

    init: Callable
    step: Callable
    extract: Callable


# all arguments are positional-only because we will rename arguments a lot
def decompose_fori_loop(lower, upper, Av, v0, /, alg: DecompAlg):
    r"""Decompose a matrix purely based on matvec-products with A.

    The semantics of this function are practically equivalent to

    ```python
    def decompose(lower, upper, Av, v0, alg):
        state = alg.init(v0)
        for _ in range(lower, upper):
            state = alg.step(state, Av)
        return alg.extract(state)
    ```

    but the implementation uses JAX' fori_loop.
    """
    init_val = alg.init(v0)

    def body_fun(_, s):
        return alg.step(s, Av=Av)

    result = control_flow.fori_loop(lower, upper, body_fun=body_fun, init_val=init_val)
    return alg.extract(result)


def lanczos(depth, /) -> DecompAlg:
    r"""Lanczos' algorithm with pre-allocation and re-orthogonalisation.

    Decompose a matrix into a product orthogonal-tridiagonal-orthogonal matrix.

    More specifically, Lanczos' algorithm orthogonally projects the original
    matrix onto the (n+1)-deep Krylov subspace

    \{ v, Av, A^2v, ..., A^n v \}

    using the Gram-Schmidt procedure.

    The spectrum of the resulting tridiagonal matrix
    approximates the spectrum of the original matrix.
    (More specifically, the spectrum tends to the 'most extreme' eigenvalues.)
    """
    # this algorithm is massively unstable.
    # but despite this instability, quadrature might be stable?
    # https://www.sciencedirect.com/science/article/abs/pii/S0920563200918164
    return DecompAlg(
        init=func.partial(_lanczos_init, depth),
        step=_lanczos_apply,
        extract=_lanczos_extract,
    )


class _LanczosState(containers.NamedTuple):
    i: int
    basis: Any
    tridiag: Any
    q: Any


def _lanczos_init(depth: int, init_vec: Array) -> _LanczosState:
    (ncols,) = np.shape(init_vec)
    if depth >= ncols or depth < 1:
        raise ValueError

    (ncols,) = init_vec.shape
    diag = np.zeros((depth + 1,))
    offdiag = np.zeros((depth,))
    basis = np.zeros((depth + 1, ncols))

    return _LanczosState(0, basis, (diag, offdiag), init_vec)


def _lanczos_apply(state: _LanczosState, Av: Callable) -> _LanczosState:
    i, basis, (diag, offdiag), vec = state

    # This one is a hack:
    #
    # Re-orthogonalise against ALL basis elements before storing.
    # Note: we re-orthogonalise against ALL columns of Q, not just
    # the ones we have already computed. This increases the complexity
    # of the whole iteration from n(n+1)/2 to n^2, but has the advantage
    # that the whole computation has static bounds (thus we can JIT it all).
    # Since 'Q' is padded with zeros, the numerical values are identical
    # between both modes of computing.
    vec, length = _normalise(vec)
    vec, _ = _gram_schmidt_orthogonalise_set(vec, basis)

    # I don't know why, but this re-normalisation is soooo crucial
    vec, _ = _normalise(vec)
    basis = basis.at[i, :].set(vec)

    # When i==0, Q[i-1] is Q[-1] and again, we benefit from the fact
    #  that Q is initialised with zeros.
    vec = Av(vec)
    basis_vectors_previous = np.asarray([basis[i], basis[i - 1]])
    vec, (coeff, _) = _gram_schmidt_orthogonalise_set(vec, basis_vectors_previous)
    diag = diag.at[i].set(coeff)
    offdiag = offdiag.at[i - 1].set(length)

    return _LanczosState(i + 1, basis, (diag, offdiag), vec)


def _lanczos_extract(state: _LanczosState, /):
    _, basis, (diag, offdiag), _ = state
    return basis, (diag, offdiag)


def _normalise(vec):
    length = linalg.vector_norm(vec)
    return vec / length, length


def _gram_schmidt_orthogonalise_set(vec, vectors):  # Gram-Schmidt
    vec, coeffs = control_flow.scan(_gram_schmidt_orthogonalise, vec, xs=vectors)
    return vec, coeffs


def _gram_schmidt_orthogonalise(vec1, vec2):
    coeff = linalg.vecdot(vec1, vec2)
    vec_ortho = vec1 - coeff * vec2
    return vec_ortho, coeff



"""Numerical linear algebra."""

import jax.numpy as jnp


def vector_norm(x, /):
    return jnp.linalg.norm(x)


def matrix_norm(x, /):
    return jnp.linalg.norm(x)


def qr(x, /, *, mode="reduced"):
    return jnp.linalg.qr(x, mode=mode)


def eigh(x, /):
    return jnp.linalg.eigh(x)


def slogdet(x, /):
    return jnp.linalg.slogdet(x)


def vecdot(x1, x2, /):
    return jnp.dot(x1, x2)


def diagonal(x, /, offset=0):
    """Extract the diagonal of a matrix."""
    return jnp.diag(x, offset)


def diagonal_matrix(x, /, offset=0):  # not part of array API
    """Construct a diagonal matrix."""
    return jnp.diag(x, offset)


def trace(x, /):
    return jnp.trace(x)



"""Hutchinson-style trace and diagonal estimation."""


from matfree import montecarlo
from matfree.backend import containers, control_flow, func, linalg, np, prng
from matfree.backend.typing import Any, Array, Callable


def trace(Av: Callable, /, **kwargs) -> Array:
    """Estimate the trace of a matrix stochastically.

    Parameters
    ----------
    Av:
        Matrix-vector product function.
    **kwargs:
        Keyword-arguments to be passed to
        [montecarlo.estimate()][matfree.montecarlo.estimate].
    """

    def quadform(vec):
        return linalg.vecdot(vec, Av(vec))

    return montecarlo.estimate(quadform, **kwargs)


def frobeniusnorm_squared(Av: Callable, /, **kwargs) -> Array:
    r"""Estimate the squared Frobenius norm of a matrix stochastically.

    The Frobenius norm of a matrix $A$ is defined as

    $$
    \|A\|_F^2 = \text{trace}(A^\top A)
    $$

    so computing squared Frobenius norms amounts to trace estimation.

    Parameters
    ----------
    Av:
        Matrix-vector product function.
    **kwargs:
        Keyword-arguments to be passed to
        [montecarlo.estimate()][matfree.montecarlo.estimate].

    """

    def quadform(vec):
        x = Av(vec)
        return linalg.vecdot(x, x)

    return montecarlo.estimate(quadform, **kwargs)


def diagonal_with_control_variate(Av: Callable, control: Array, /, **kwargs) -> Array:
    """Estimate the diagonal of a matrix stochastically and with a control variate.

    Parameters
    ----------
    Av:
        Matrix-vector product function.
    control:
        Control variate.
        This should be the best-possible estimate of the diagonal of the matrix.
    **kwargs:
        Keyword-arguments to be passed to
        [montecarlo.estimate()][matfree.montecarlo.estimate].

    """
    return diagonal(lambda v: Av(v) - control * v, **kwargs) + control


def diagonal(Av: Callable, /, **kwargs) -> Array:
    """Estimate the diagonal of a matrix stochastically.

    Parameters
    ----------
    Av:
        Matrix-vector product function.
    **kwargs:
        Keyword-arguments to be passed to
        [montecarlo.estimate()][matfree.montecarlo.estimate].

    """

    def quadform(vec):
        return vec * Av(vec)

    return montecarlo.estimate(quadform, **kwargs)


def trace_and_diagonal(Av: Callable, /, *, sample_fun: Callable, key: Array, **kwargs):
    """Jointly estimate the trace and the diagonal stochastically.

    The advantage of computing both quantities simultaneously is
    that the diagonal estimate
    may serve as a control variate for the trace estimate,
    thus reducing the variance of the estimator
    (and thereby accelerating convergence.)

    Parameters
    ----------
    Av:
        Matrix-vector product function.
    sample_fun:
        Sampling function.
        Usually, either [montecarlo.normal][matfree.montecarlo.normal]
        or [montecarlo.rademacher][matfree.montecarlo.normal].
    key:
        Pseudo-random number generator key.
    **kwargs:
        Keyword-arguments to be passed to
        [diagonal_multilevel()][matfree.hutch.diagonal_multilevel].


    See:
    Adams et al., Estimating the Spectral Density of Large Implicit Matrices, 2018.
    """
    fx_value = func.eval_shape(sample_fun, key)
    init = np.zeros(shape=fx_value.shape, dtype=fx_value.dtype)
    final = diagonal_multilevel(Av, init, sample_fun=sample_fun, key=key, **kwargs)
    return np.sum(final), final


class _EstState(containers.NamedTuple):
    diagonal_estimate: Any
    key: Any


def diagonal_multilevel(
    Av: Callable,
    init: Array,
    /,
    *,
    key: Array,
    sample_fun: Callable,
    num_levels: int,
    num_batches_per_level: int = 1,
    num_samples_per_batch: int = 1,
) -> Array:
    """Estimate the diagonal in a multilevel framework.

    The general idea is that a diagonal estimate serves as a control variate
    for the next step's diagonal estimate.


    Parameters
    ----------
    Av:
        Matrix-vector product function.
    init:
        Initial guess.
    key:
        Pseudo-random number generator key.
    sample_fun:
        Sampling function.
        Usually, either [montecarlo.normal][matfree.montecarlo.normal]
        or [montecarlo.rademacher][matfree.montecarlo.normal].
    num_levels:
        Number of levels.
    num_batches_per_level:
        Number of batches per level.
    num_samples_per_batch:
        Number of samples per batch (per level).

    """
    kwargs = {
        "sample_fun": sample_fun,
        "num_batches": num_batches_per_level,
        "num_samples_per_batch": num_samples_per_batch,
    }

    def update_fun(level: int, x: _EstState) -> _EstState:
        """Update the diagonal estimate."""
        diag, k = x

        _, subkey = prng.split(k, num=2)
        update = diagonal_with_control_variate(Av, diag, key=subkey, **kwargs)

        diag = _incr(diag, level, update)
        return _EstState(diag, subkey)

    state = _EstState(diagonal_estimate=init, key=key)
    state = control_flow.fori_loop(0, num_levels, body_fun=update_fun, init_val=state)
    (final, *_) = state
    return final


def _incr(old, count, incoming):
    return (old * count + incoming) / (count + 1)



"""Tests for basic trace estimators."""

from matfree import hutch, montecarlo
from matfree.backend import func, linalg, np, prng, testing


@testing.fixture(name="fun")
def fixture_fun():
    """Create a nonlinear, to-be-differentiated function."""

    def f(x):
        return np.sin(np.flip(np.cos(x)) + 1.0) * np.sin(x) + 1.0

    return f


@testing.fixture(name="key")
def fixture_key():
    """Fix a pseudo-random number generator."""
    return prng.prng_key(seed=1)


@testing.parametrize("num_batches", [1_000])
@testing.parametrize("num_samples_per_batch", [1_000])
@testing.parametrize("dim", [1, 10])
@testing.parametrize("sample_fun", [montecarlo.normal, montecarlo.rademacher])
def test_frobeniusnorm_squared(
    fun, key, num_batches, num_samples_per_batch, dim, sample_fun
):
    """Assert that the Frobenius norm estimate is accurate."""
    # Linearise function
    x0 = prng.uniform(key, shape=(dim,))  # random lin. point
    _, jvp = func.linearize(fun, x0)
    J = func.jacfwd(fun)(x0)

    # Estimate the trace
    fun = sample_fun(shape=np.shape(x0), dtype=np.dtype(x0))
    estimate = hutch.frobeniusnorm_squared(
        jvp,
        num_batches=num_batches,
        key=key,
        num_samples_per_batch=num_samples_per_batch,
        sample_fun=fun,
    )
    truth = linalg.trace(J.T @ J)
    assert np.allclose(estimate, truth, rtol=1e-2)


@testing.parametrize("num_batches", [1_000])
@testing.parametrize("num_samples_per_batch", [1_000])
@testing.parametrize("dim", [1, 10])
@testing.parametrize("sample_fun", [montecarlo.normal, montecarlo.rademacher])
def test_trace(fun, key, num_batches, num_samples_per_batch, dim, sample_fun):
    """Assert that the estimated trace approximates the true trace accurately."""
    # Linearise function
    x0 = prng.uniform(key, shape=(dim,))  # random lin. point
    _, jvp = func.linearize(fun, x0)
    J = func.jacfwd(fun)(x0)

    # Estimate the trace
    fun = sample_fun(shape=np.shape(x0), dtype=np.dtype(x0))
    estimate = hutch.trace(
        jvp,
        num_batches=num_batches,
        key=key,
        num_samples_per_batch=num_samples_per_batch,
        sample_fun=fun,
    )
    truth = linalg.trace(J)
    assert np.allclose(estimate, truth, rtol=1e-2)


@testing.parametrize("num_batches", [1_000])
@testing.parametrize("num_samples_per_batch", [1_000])
@testing.parametrize("dim", [1, 10])
@testing.parametrize("sample_fun", [montecarlo.normal, montecarlo.rademacher])
def test_diagonal(fun, key, num_batches, num_samples_per_batch, dim, sample_fun):
    """Assert that the estimated diagonal approximates the true diagonal accurately."""
    # Linearise function
    x0 = prng.uniform(key, shape=(dim,))  # random lin. point
    _, jvp = func.linearize(fun, x0)
    J = func.jacfwd(fun)(x0)

    # Estimate the trace
    fun = sample_fun(shape=np.shape(x0), dtype=np.dtype(x0))
    estimate = hutch.diagonal(
        jvp,
        num_batches=num_batches,
        key=key,
        num_samples_per_batch=num_samples_per_batch,
        sample_fun=fun,
    )
    truth = linalg.diagonal(J)
    assert np.allclose(estimate, truth, rtol=1e-2)


@testing.parametrize("num_samples", [10_000])
@testing.parametrize("dim", [5])
@testing.parametrize("sample_fun", [montecarlo.normal, montecarlo.rademacher])
def test_trace_and_diagonal(fun, key, num_samples, dim, sample_fun):
    """Assert that the estimated trace and diagonal approximations are accurate."""
    # Linearise function
    x0 = prng.uniform(key, shape=(dim,))
    _, jvp = func.linearize(fun, x0)
    J = func.jacfwd(fun)(x0)

    # Estimate the trace
    fun = sample_fun(shape=np.shape(x0), dtype=np.dtype(x0))
    trace, diag = hutch.trace_and_diagonal(
        jvp, key=key, num_levels=num_samples, sample_fun=fun
    )

    # Print errors if test fails
    error_diag = linalg.vector_norm(diag - linalg.diagonal(J))
    error_trace = linalg.vector_norm(trace - linalg.trace(J))
    assert np.allclose(diag, linalg.diagonal(J), rtol=1e-2), error_diag
    assert np.allclose(trace, linalg.trace(J), rtol=1e-2), error_trace



"""Tests for Monte-Carlo machinery."""

from matfree import montecarlo
from matfree.backend import np, prng, testing


def test_van_der_corput():
    """Assert that the van-der-Corput sequence yields values as expected."""
    expected = np.asarray([0, 0.5, 0.25, 0.75, 0.125, 0.625, 0.375, 0.875, 0.0625])
    received = np.asarray([montecarlo.van_der_corput(i) for i in range(9)])
    assert np.allclose(received, expected)

    expected = np.asarray([0.0, 1 / 3, 2 / 3, 1 / 9, 4 / 9, 7 / 9, 2 / 9, 5 / 9, 8 / 9])
    received = np.asarray([montecarlo.van_der_corput(i, base=3) for i in range(9)])
    assert np.allclose(received, expected)


@testing.parametrize("key", [prng.prng_key(1)])
@testing.parametrize("num_batches, num_samples", [[1, 10_000], [10_000, 1], [100, 100]])
def test_mean(key, num_batches, num_samples):
    """Assert that the mean estimate is accurate."""

    def fun(x):
        return x**2

    received = montecarlo.estimate(
        fun,
        num_batches=num_batches,
        num_samples_per_batch=num_samples,
        key=key,
        sample_fun=montecarlo.normal(shape=()),
    )
    assert np.allclose(received, 1.0, rtol=1e-1)



"""Test utilities."""

from matfree.backend import linalg, np


def generate_symmetric_matrix_from_eigvals(eigvals, /):
    """Generate a symmetric matrix with prescribed eigenvalues."""
    (n,) = eigvals.shape

    # Need _some_ matrix to start with
    A = np.reshape(np.arange(1.0, n**2 + 1.0), (n, n))
    A = A / linalg.matrix_norm(A)
    X = A.T @ A + np.eye(n)

    # QR decompose. We need the orthogonal matrix.
    # Treat Q as a stack of eigenvectors.
    Q, R = linalg.qr(X)

    # Treat Q as eigenvectors, and 'D' as eigenvalues.
    # return Q D Q.T.
    # This matrix will be dense, symmetric, and have a given spectrum.
    return Q @ (eigvals[:, None] * Q.T)



"""Tests for Lanczos functionality."""

from matfree import decomp, test_util
from matfree.backend import linalg, np, prng, testing


@testing.fixture()
def A(n, num_significant_eigvals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    d = np.arange(n) + 10.0
    d = d.at[num_significant_eigvals:].set(0.001)

    return test_util.generate_symmetric_matrix_from_eigvals(d)


@testing.parametrize("n", [6])
@testing.parametrize("num_significant_eigvals", [4])
def test_tridiagonal_error_for_too_high_order(A):
    """Assert graceful failure if the depth matches or exceeds the number of columns."""
    n, _ = np.shape(A)
    key = prng.prng_key(1)
    v0 = prng.normal(key, shape=(n,))
    with testing.raises(ValueError):
        alg = decomp.lanczos(n + 10)
        _ = decomp.decompose_fori_loop(0, n + 10, lambda v: A @ v, v0, alg=alg)
    with testing.raises(ValueError):
        alg = decomp.lanczos(n)
        _ = decomp.decompose_fori_loop(0, n + 1, lambda v: A @ v, v0, alg=alg)


@testing.parametrize("n", [6])
@testing.parametrize("num_significant_eigvals", [6])
def test_tridiagonal_max_order(A):
    """If m == n, the matrix should be equal to the full tridiagonal."""
    n, _ = np.shape(A)
    order = n - 1
    key = prng.prng_key(1)
    v0 = prng.normal(key, shape=(n,))
    alg = decomp.lanczos(order)
    Q, (d_m, e_m) = decomp.decompose_fori_loop(
        0, order + 1, lambda v: A @ v, v0, alg=alg
    )

    # Lanczos is not stable.
    tols_decomp = {"atol": 1e-5, "rtol": 1e-5}

    # Since full-order mode: Q must be unitary
    assert np.shape(Q) == (order + 1, n)
    assert np.allclose(Q @ Q.T, np.eye(n), **tols_decomp), Q @ Q.T
    assert np.allclose(Q.T @ Q, np.eye(n), **tols_decomp), Q.T @ Q

    # T = Q A Qt
    T = _sym_tridiagonal_dense(d_m, e_m)
    QAQt = Q @ A @ Q.T
    assert np.shape(T) == (order + 1, order + 1)

    # Fail early if the (off)diagonals don't coincide
    assert np.allclose(linalg.diagonal(QAQt), d_m, **tols_decomp)
    assert np.allclose(linalg.diagonal(QAQt, 1), e_m, **tols_decomp)
    assert np.allclose(linalg.diagonal(QAQt, -1), e_m, **tols_decomp)

    # Test the full decomposition
    # (i.e. assert that the off-tridiagonal elements are actually small)
    # be loose with this test. off-diagonal elements accumulate quickly.
    tols_decomp = {"atol": 1e-5, "rtol": 1e-5}
    assert np.allclose(QAQt, T, **tols_decomp)

    # Since full-order mode: Qt T Q = A
    # Since Q is unitary and T = Q A Qt, this test
    # should always pass.
    assert np.allclose(Q.T @ T @ Q, A, **tols_decomp)


@testing.parametrize("n", [50])
@testing.parametrize("num_significant_eigvals", [4])
@testing.parametrize("order", [6])  # ~1.5 * num_significant_eigvals
def test_tridiagonal(A, order):
    """Test that Lanczos tridiagonalisation yields an orthogonal-tridiagonal decomp."""
    n, _ = np.shape(A)
    key = prng.prng_key(1)
    v0 = prng.normal(key, shape=(n,))
    alg = decomp.lanczos(order)
    Q, tridiag = decomp.decompose_fori_loop(0, order + 1, lambda v: A @ v, v0, alg=alg)
    (d_m, e_m) = tridiag

    # Lanczos is not stable.
    tols_decomp = {"atol": 1e-5, "rtol": 1e-5}

    assert np.shape(Q) == (order + 1, n)
    assert np.allclose(Q @ Q.T, np.eye(order + 1), **tols_decomp), Q @ Q.T

    # T = Q A Qt
    T = _sym_tridiagonal_dense(d_m, e_m)
    QAQt = Q @ A @ Q.T
    assert np.shape(T) == (order + 1, order + 1)

    # Fail early if the (off)diagonals don't coincide
    assert np.allclose(linalg.diagonal(QAQt), d_m, **tols_decomp)
    assert np.allclose(linalg.diagonal(QAQt, 1), e_m, **tols_decomp)
    assert np.allclose(linalg.diagonal(QAQt, -1), e_m, **tols_decomp)

    # Test the full decompoisition
    assert np.allclose(QAQt, T, **tols_decomp)


def _sym_tridiagonal_dense(d, e):
    diag = linalg.diagonal_matrix(d)
    offdiag1 = linalg.diagonal_matrix(e, 1)
    offdiag2 = linalg.diagonal_matrix(e, -1)
    return diag + offdiag1 + offdiag2



"""Tests for (selected) autodiff functionality."""


from matfree import montecarlo, slq, test_util
from matfree.backend import np, prng, testing


@testing.fixture()
def A(n, num_significant_eigvals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    d = np.arange(n) + 10.0
    d = d.at[num_significant_eigvals:].set(0.001)

    return test_util.generate_symmetric_matrix_from_eigvals(d)


@testing.parametrize("n", [200])
@testing.parametrize("num_significant_eigvals", [30])
@testing.parametrize("order", [10])
# usually: ~1.5 * num_significant_eigvals.
# But logdet seems to converge sooo much faster.
def test_logdet(A, order):
    """Assert that log-determinant computation admits valid VJPs and JVPs."""
    key = prng.prng_key(1)

    def fun(s):
        return _logdet(s, order, key)

    testing.check_grads(fun, (A,), order=1, atol=1e-1, rtol=1e-1)


def _logdet(A, order, key):
    n, _ = np.shape(A)
    fun = montecarlo.normal(shape=(n,))
    return slq.logdet(
        lambda v: A @ v,
        order,
        key=key,
        num_samples_per_batch=10,
        num_batches=1,
        sample_fun=fun,
    )
